import contextlib
import json
import typing
from collections import defaultdict

from django.contrib.contenttypes.models import ContentType
from django.db import IntegrityError, transaction

from django_gem.core import get_model_natural_key
from django_gem.logger import logger
from django_gem.models.mixins import GemModelMixin
from django_gem.toolkit import forge, gem_settings
from django_gem.toolkit.chest import SealedChestItem


class Saw:
    @classmethod
    def _get_queryset(cls, model_class, object_ids):
        from django_gem.toolkit import cutter_registry

        queryset = model_class.objects.filter(id__in=object_ids)
        select_related_fields, prefetch_related_fields = [], []
        cutter_model = cutter_registry.get_cutter_for_model(model_class)
        if cutter_model:
            (select_related_fields, prefetch_related_fields) = (
                cutter_model.select_related_fields,
                cutter_model.prefetch_related_fields,
            )
        if select_related_fields:
            queryset = queryset.select_related(*select_related_fields)
        if prefetch_related_fields:
            queryset = queryset.prefetch_related(*prefetch_related_fields)
        return queryset

    @classmethod
    def cut_content_type(cls, natural_key: str, field_names: list):
        gem_cutting_enabled = gem_settings.GEM_CUTTING_ENABLED

        if not gem_cutting_enabled:
            return
        try:
            content_type: ContentType = ContentType.objects.get_by_natural_key(
                *natural_key.split(".")
            )
        except ContentType.DoesNotExist:
            return

        model_class = content_type.model_class()
        for model_instance in model_class.objects.all():
            if isinstance(model_class, GemModelMixin):
                model_class.update_cutting_started_at([model_instance.id])
            forge.cut_model_fields(model_instance, field_names)
            if isinstance(model_class, GemModelMixin):
                model_class.update_cutting_completed_at([model_instance.id])

    @classmethod
    def cut_models(cls, sealed_chest: str, sealed_sources: str):
        gem_cutting_enabled = gem_settings.GEM_CUTTING_ENABLED

        cut_mapping = defaultdict(int)

        if not gem_cutting_enabled:
            return
        try:
            unsealed_chest: typing.Dict[str, typing.List[typing.Dict]] = json.loads(sealed_chest)
        except json.JSONDecodeError:
            return

        logger.info(f"Cutting initiated by: {sealed_sources}; for: {sealed_chest}")

        for natural_key, chest_items in unsealed_chest.items():
            try:
                content_type: ContentType = ContentType.objects.get_by_natural_key(
                    *natural_key.split(".")
                )
            except ContentType.DoesNotExist:
                continue

            sealed_chest_mapping: typing.Dict[str, SealedChestItem] = {
                SealedChestItem(**item).object_id: SealedChestItem(**item) for item in chest_items
            }
            chest_item_object_ids = [chest_item for chest_item in sealed_chest_mapping]

            model_class = content_type.model_class()

            if isinstance(model_class, GemModelMixin):
                model_class.update_cutting_started_at(chest_item_object_ids)

            queryset = cls._get_queryset(model_class, chest_item_object_ids)

            for model_instance in queryset:
                model_instance_id = str(model_instance.id)
                if (sealed_chest_item := sealed_chest_mapping.get(str(model_instance_id))) is None:
                    continue
                with contextlib.suppress(IntegrityError):
                    # Because of the distributed nature of cutting, the update to gem can happen
                    # after the referencing model was deleted. This will ensure we don't stop
                    # all the cuttings if something goes wrong.
                    forge.cut_model_fields(model_instance, sealed_chest_item.gem_fields)
                    cut_mapping[natural_key] += 1

            if isinstance(model_class, GemModelMixin):
                model_class.update_cutting_completed_at(chest_item_object_ids)

        cut_mapping = dict(cut_mapping)
        logger.info(f"Cutting finished, cutting results: {cut_mapping}")

        return cut_mapping

    @classmethod
    def cut_queryset(cls, natural_key: str, object_ids: list):
        gem_cutting_enabled = gem_settings.GEM_CUTTING_ENABLED

        if not gem_cutting_enabled:
            return
        from django_gem.toolkit import forge, smith

        try:
            content_type: ContentType = ContentType.objects.get_by_natural_key(
                *natural_key.split(".")
            )
        except ContentType.DoesNotExist:
            return

        cut_mapping = defaultdict(int)

        model_class = content_type.model_class()

        queryset = cls._get_queryset(model_class, object_ids)

        for model_instance in queryset:
            reverse_cutter_model = forge.cutter_registry.get_reverse_cutter_for_model(
                model_instance
            )
            if not reverse_cutter_model or not reverse_cutter_model.model:
                continue

            for reverse_cutter_trigger in reverse_cutter_model.triggers:
                object_ids = list(reverse_cutter_trigger.callback(model_instance))
                if not object_ids:
                    continue
                natural_key = get_model_natural_key(reverse_cutter_trigger.model)
                smith.add_item(
                    natural_key=natural_key,
                    object_ids=object_ids,
                    gem_fields=reverse_cutter_trigger.gem_fields,
                )
                cut_mapping[natural_key] += 1
            with transaction.atomic():
                smith.initiate_cutting()
        cut_mapping = dict(cut_mapping)
        return cut_mapping



import typing
from collections import defaultdict
from dataclasses import asdict, dataclass

from django_gem.entities.source import CutSource


@dataclass
class ChestItem:
    gem_fields: list
    object_ids: list


@dataclass
class SealedChestItem:
    object_id: str
    gem_fields: list


class Chest:
    """Storage for all pending cuttings"""

    cut_models: typing.Dict[str, typing.List[ChestItem]] = defaultdict(list)
    cut_sources: typing.Dict[str, typing.List[CutSource]] = defaultdict(list)

    @classmethod
    def seal_items(cls, chest_items: typing.List[ChestItem]) -> typing.List[SealedChestItem]:
        sealed_item_mapping: typing.Dict[str, SealedChestItem] = {}
        for chest_item in chest_items:
            for object_id in chest_item.object_ids:
                if sealed_item_mapping.get(object_id):
                    sealed_item_mapping[object_id].gem_fields = list(
                        {*sealed_item_mapping[object_id].gem_fields, *chest_item.gem_fields}
                    )
                else:
                    sealed_item_mapping[object_id] = SealedChestItem(
                        object_id=object_id,
                        gem_fields=chest_item.gem_fields,
                    )
        return list(sealed_item_mapping.values())

    def add_chest_item(self, natural_key: str, object_ids: list, gem_fields: list):
        object_ids = [object_id for object_id in object_ids if object_id is not None]
        if not object_ids:
            return

        self.cut_models[natural_key].append(
            ChestItem(
                gem_fields=gem_fields,
                object_ids=[str(object_id) for object_id in object_ids if object_id is not None],
            )
        )

    def add_cut_source(self, natural_key: str, cut_source: CutSource):
        self.cut_sources[natural_key].append(cut_source)

    def reset(self):
        self.cut_models = defaultdict(list)
        self.cut_sources = defaultdict(list)

    def is_empty(self):
        return not bool(self.cut_models)

    def get_sealed_chest(self):
        sealed_chest = {}

        for natural_key, chest_items in self.cut_models.items():
            sealed_chest[natural_key] = [
                asdict(item) for item in self.seal_items(chest_items=chest_items)
            ]
        return sealed_chest

    def get_sealed_sources(self):
        sealed_sources = {}
        for natural_key, source_items in self.cut_sources.items():
            sealed_sources[natural_key] = [asdict(item) for item in source_items]
        return sealed_sources



import json

from django_gem.tasks.cut_models import cut_models_task
from django_gem.toolkit.anvils.base import BaseAnvil


class CeleryAnvil(BaseAnvil):
    @classmethod
    def cut(cls, sealed_chest, sealed_sources):
        cut_models_task.delay(
            sealed_chest=json.dumps(sealed_chest), sealed_sources=json.dumps(sealed_sources)
        )



import typing

from dirtyfields import DirtyFieldsMixin
from django.db import transaction

from django_gem.constants import GemCuttingMode
from django_gem.core import get_model_natural_key
from django_gem.entities.context import gem_cutting_context
from django_gem.entities.registry import CutterBatchItem
from django_gem.entities.source import CutSource
from django_gem.models.mixins import GemCutConditionalMixin
from django_gem.overrides.utils import add_cut_items, is_cutting_needed


def save_cutting_hook(
    save_func,
    cut_batch_items: typing.List[CutterBatchItem],
):
    def wrapped_trigger(self, *args, **kwargs):
        from django_gem.toolkit import gem_settings, smith

        gem_cutting_enabled = gem_settings.GEM_CUTTING_ENABLED

        should_cut = True
        if isinstance(self, GemCutConditionalMixin):
            should_cut = self.should_cut_on_save(*args, **kwargs)

        with transaction.atomic():
            prepared_cut_batch_items = []
            changed_cut_batch_items = []
            if should_cut:
                prepared_cut_batch_items = [
                    cut_batch_item
                    for cut_batch_item in cut_batch_items
                    if is_cutting_needed(self, cut_batch_item.affected_fields)
                ]

            save_func(self, *args, **kwargs)
            if not gem_cutting_enabled or not should_cut:
                return

            for cut_batch_item in prepared_cut_batch_items:
                cut_batch_item.load_object_ids(self)
                changed_cut_batch_items.append(cut_batch_item)

            add_cut_items(smith, changed_cut_batch_items)
            affected_fields = []
            if isinstance(self, DirtyFieldsMixin):
                dirty_fields = self.get_dirty_fields(check_relationship=True)
                if dirty_fields:
                    affected_fields = list(dirty_fields.keys())

            smith.add_source(
                get_model_natural_key(self),  # noqa
                CutSource(
                    object_id=str(self.id),  # noqa
                    affected_fields=affected_fields,
                ),
            )
            if gem_cutting_context.cutting_mode == GemCuttingMode.TRANSACTION:
                transaction.on_commit(
                    lambda: smith.initiate_cutting(),
                )

    return wrapped_trigger



import typing
from dataclasses import dataclass, field

from django.db import models

from django_gem.constants import CallbackType


@dataclass
class CutterSideEffect:
    model: typing.Type[models.Model]
    callback: CallbackType
    affected_fields: typing.List[str] = field(default_factory=lambda: [])
    gem_fields: typing.List[str] = field(default_factory=lambda: [])

    @property
    def callback_group(self) -> str:
        return self.callback.__name__


@dataclass
class CutterRelatedGemSideEffect:
    related_gem_model: typing.Type[models.Model]
    callback: CallbackType
    self_affected_fields: typing.List[str] = field(default_factory=lambda: [])
    gem_fields: typing.List[str] = field(default_factory=lambda: [])



import decimal

from crum import get_current_user
from dirtyfields import DirtyFieldsMixin
from django.db import models
from django.utils import timezone

from django_gem.constants import GemLogEntryAction
from django_gem.models import GemLogEntry


class GemTriggerMixin(DirtyFieldsMixin):
    def is_any_field_changed(self, fields: list):
        dirty_fields = self.get_dirty_fields(check_relationship=True)
        if not dirty_fields:
            return False  # No fields changed - we don't recalculate
        return any(field in dirty_fields for field in fields)


class GemAuditLogMixin(DirtyFieldsMixin):
    @classmethod
    def field_value(cls, instance, field_name):
        if not instance:
            return None
        field_value = instance.__getattribute__(field_name)
        if isinstance(field_value, decimal.Decimal):
            return float(field_value)
        if isinstance(field_value, models.Model):
            return field_value.pk
        return field_value

    def save_log_entry(self, is_adding: bool, dirty_fields: dict):
        if not hasattr(self, "get_gem_fields") or not dirty_fields:
            return
        changes = {
            field_name: self.field_value(self, field_name) for field_name in self.get_gem_fields()
        }

        if changes:
            user = get_current_user()
            if user and not user.pk:
                user = None
            GemLogEntry.objects.log_create(
                self,
                action=GemLogEntryAction.CREATE if is_adding else GemLogEntryAction.UPDATE,
                changes=changes,
                actor=user,
            )

    def save(self, *args, **kwargs):
        is_adding, dirty_fields = self._state.adding, self.get_dirty_fields()  # noqa
        super().save(*args, **kwargs)  # noqa
        self.save_log_entry(is_adding, dirty_fields)

    class Meta:
        audit_log_fields = []


class GemCutConditionalMixin:
    def should_cut_on_save(self, *args, **kwargs):  # noqa
        return True

    def should_cut_on_delete(self, *args, **kwargs):  # noqa
        return True


class GemModelMixin(models.Model):
    cutting_started_at = models.DateTimeField(null=True, blank=True)
    cutting_completed_at = models.DateTimeField(null=True, blank=True)

    @classmethod
    def update_cutting_started_at(cls, ids: list):
        cls.objects.filter(id__in=ids).update(cutting_started_at=timezone.now())

    @classmethod
    def update_cutting_completed_at(cls, ids: list):
        cls.objects.filter(id__in=ids).update(cutting_completed_at=timezone.now())

    @property
    def is_cutting_in_progress(self):
        return (
            self.cutting_started_at
            and self.cutting_completed_at
            and self.cutting_started_at > self.cutting_completed_at
        )

    class Meta:
        abstract = True



from django_gem.toolkit.settings import gem_settings  # isort:skip
from django_gem.toolkit.registry import cutter_registry  # isort:skip
from django_gem.toolkit.forge import forge  # isort:skip
from django_gem.toolkit.smith import smith  # isort:skip



import json

from celery import shared_task


@shared_task(name="cut_models_task")
def cut_models_task(sealed_chest: str, sealed_sources: str):
    from django_gem.toolkit.saw import Saw

    cut_mapping = json.dumps(Saw.cut_models(sealed_chest, sealed_sources))
    return cut_mapping



from django_gem.entities.context import gem_cutting_context


class CuttingAlwaysEager:
    def __enter__(self):
        gem_cutting_context.return_eager_value = True

    def __exit__(self, exc_type, exc_value, exc_traceback):
        gem_cutting_context.return_eager_value = False



import importlib
import typing

from django_gem.entities.source import CutSource
from django_gem.logger import logger
from django_gem.toolkit import gem_settings
from django_gem.toolkit.anvils.base import BaseAnvil
from django_gem.toolkit.chest import Chest


class Smith:
    """Handles data collection and initiates cut process"""

    chest = Chest()
    anvils: typing.List[BaseAnvil] = []

    def add_item(self, natural_key: str, object_ids: list, gem_fields: list):
        self.chest.add_chest_item(natural_key, object_ids, gem_fields)

    def add_source(self, natural_key: str, source: CutSource):
        self.chest.add_cut_source(natural_key, source)

    def add_anvil(self, anvil: BaseAnvil):
        self.anvils.append(anvil)

    def load(self):
        for anvil_import in gem_settings.GEM_ANVILS:
            try:
                anvil_module, anvil_class = anvil_import.rsplit(".", 1)
            except ValueError:
                logger.error("Error importing %s: Anvil import misconfigured", anvil_import)
                continue
            try:
                anvil: typing.Type[BaseAnvil] = getattr(
                    importlib.import_module(anvil_module), anvil_class
                )
            except ImportError as e:
                logger.error("Error importing %s: %s", anvil_import, e)
                continue
            if not anvil:
                logger.error("Error configuring %s: %s not found", anvil_import, anvil_class)
                continue
            self.add_anvil(anvil())

    def initiate_cutting(self):
        if self.chest.is_empty():
            return

        sealed_chest = self.chest.get_sealed_chest()
        sealed_sources = self.chest.get_sealed_sources()
        for anvil in self.anvils:
            anvil.cut(sealed_chest, sealed_sources)
        self.chest.reset()


smith = Smith()



import typing

from django.db import transaction

from django_gem.constants import GemCuttingMode
from django_gem.core import get_model_natural_key
from django_gem.entities.context import gem_cutting_context
from django_gem.entities.registry import CutterBatchItem
from django_gem.entities.source import CutSource
from django_gem.models.mixins import GemCutConditionalMixin
from django_gem.overrides.utils import add_cut_items


def delete_cutting_hook(
    delete_func,
    cut_batch_items: typing.List[CutterBatchItem],
):
    def wrapped_trigger(self, *args, **kwargs):
        from django_gem.toolkit import gem_settings, smith

        gem_cutting_enabled = gem_settings.GEM_CUTTING_ENABLED

        should_cut = True
        if isinstance(self, GemCutConditionalMixin):
            should_cut = self.should_cut_on_delete(*args, **kwargs)

        if should_cut:
            for cut_batch_item in cut_batch_items:
                cut_batch_item.load_object_ids(self)

            smith.add_source(
                get_model_natural_key(self), CutSource(object_id=str(self.id), affected_fields=[])
            )

        with transaction.atomic():
            result = delete_func(self, *args, **kwargs)
            if not gem_cutting_enabled or not should_cut:
                return result

            add_cut_items(smith, cut_batch_items)
            if gem_cutting_context.cutting_mode == GemCuttingMode.TRANSACTION:
                transaction.on_commit(
                    lambda: smith.initiate_cutting(),
                )
            return result

    return wrapped_trigger



class BaseCutterMetaclass(type):
    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        for attr_name, attr_value in attrs.items():
            side_effects = (
                getattr(attr_value, "side_effects", None)
                # Supporting Django cached_property here
                or getattr(getattr(attr_value, "real_func", None), "side_effects", None)
                # Supporting functools cached_property here
                or getattr(getattr(attr_value, "func", None), "side_effects", None)
                # Supporting regular property here
                or getattr(getattr(attr_value, "fget", None), "side_effects", None)
            )
            if side_effects is not None:
                cls.side_effects = [
                    *getattr(cls, "side_effects", []),
                    *side_effects,
                ]


class BaseCutter(metaclass=BaseCutterMetaclass):
    side_effects = []

    def __init__(self, instance):
        self.instance = instance



import typing

from django.utils.translation import gettext_lazy as _

CallbackType = typing.NewType(
    "CallbackType",
    typing.Callable[[typing.Any], typing.Union[typing.List, typing.Set]],
)


class GemLogEntryAction:
    CREATE = 0
    UPDATE = 1
    DELETE = 2

    @classmethod
    def to_choices(cls):
        return (
            (cls.CREATE, _("create")),
            (cls.UPDATE, _("update")),
            (cls.DELETE, _("delete")),
        )


class GemCuttingMode:
    TRANSACTION = "transaction"
    MANUAL = "manual"



import uuid

from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from django_gem.constants import GemLogEntryAction
from django_gem.models.managers.gem_log_entry_manager import GemLogEntryManager


class GemLogEntry(models.Model):
    id = models.UUIDField(
        primary_key=True, default=uuid.uuid4, serialize=False, editable=False, unique=True
    )
    content_type = models.ForeignKey(
        to="contenttypes.ContentType",
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("content type"),
    )
    object_id = models.CharField(db_index=True, max_length=255, verbose_name=_("object id"))
    action = models.PositiveSmallIntegerField(
        choices=GemLogEntryAction.to_choices(), verbose_name=_("action"), db_index=True
    )
    actor = models.ForeignKey(
        to=settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        blank=True,
        null=True,
        related_name="+",
        verbose_name=_("actor"),
    )
    changes = models.JSONField(verbose_name=_("Model changed fields"))
    timestamp = models.DateTimeField(db_index=True, auto_now_add=True, verbose_name=_("timestamp"))

    objects = GemLogEntryManager()

    class Meta:
        get_latest_by = "timestamp"
        ordering = ["-timestamp"]
        verbose_name = _("gem log entry")
        verbose_name_plural = _("gem log entries")



from django_gem.entities.context import gem_cutting_context


class OverrideGemCuttingMode:
    def __init__(self, override_mode: str):
        self.override_mode = override_mode
        self.original_mode = gem_cutting_context.cutting_mode

    def __enter__(self):
        gem_cutting_context.cutting_mode = self.override_mode

    def __exit__(self, exc_type, exc_value, exc_traceback):
        gem_cutting_context.cutting_mode = self.original_mode



import typing

from django.utils.module_loading import autodiscover_modules

from django_gem.cutters.base import BaseCutter
from django_gem.models.base import CutterEngineBaseModel
from django_gem.toolkit import cutter_registry, gem_settings


class Forge:
    """Engine for running cut methods"""

    def __init__(self, registry):
        self.cutter_registry = registry

    def cut_model_field(self, model_instance, field_name):
        self.cut_model_fields(model_instance, [field_name])

    def cut_model_fields(self, model_instance, field_names):
        cut_fields = []
        autodiscover_modules("apps")
        cutter_model = self.cutter_registry.get_cutter_for_model(model_instance)
        if not cutter_model or not cutter_model.model:
            return None
        for registered_cutter in cutter_model.triggers:
            cutter_class, gem_class, propagate_triggers = (
                registered_cutter.cutter,
                registered_cutter.gem_class,
                registered_cutter.propagate_triggers,
            )
            cutter = cutter_class(model_instance)
            gem_instance, _ = gem_class.objects.get_or_create(
                **{gem_settings.CUTTER_MODEL_RELATED_NAME: model_instance},
            )
            for field_name in field_names:
                cutter_target_method = f"{gem_settings.CUTTER_PROPERTY_PREFIX}{field_name}"
                if not hasattr(cutter_class, cutter_target_method):
                    continue

                if field_name not in cut_fields:
                    setattr(
                        gem_instance,
                        field_name,
                        getattr(cutter, cutter_target_method, None),
                    )
                cut_fields.append(field_name)
                gem_instance = self.cut_propagated_triggers(
                    field_name,
                    propagate_triggers,
                    cutter,
                    gem_instance,
                    cut_fields,
                )
                self.cut_related_gem(gem_instance, gem_class, field_name)
            gem_instance.save()

    @classmethod
    def cut_propagated_triggers(
        cls,
        field_name: str,
        propagate_triggers,
        cutter_instance: BaseCutter,
        gem_instance,
        cut_fields: list,
        max_depth=0,
    ):
        if field_name in propagate_triggers:
            for propagated_trigger in propagate_triggers[field_name]:
                cutter_target_method = f"{gem_settings.CUTTER_PROPERTY_PREFIX}{propagated_trigger}"
                if not hasattr(cutter_instance.__class__, cutter_target_method):
                    continue

                if propagated_trigger not in cut_fields:
                    setattr(
                        gem_instance,
                        propagated_trigger,
                        getattr(cutter_instance, cutter_target_method, None),
                    )
                cut_fields.append(propagated_trigger)
                # Traversing the whole dict to find next dependencies
                if (
                    propagated_trigger != field_name
                    and max_depth < gem_settings.CUTTER_PROPAGATED_TRIGGERS_MAX_DEPTH
                ):
                    gem_instance = cls.cut_propagated_triggers(
                        propagated_trigger,
                        propagate_triggers,
                        cutter_instance,
                        gem_instance,
                        cut_fields,
                        max_depth + 1,
                    )
        return gem_instance

    def cut_related_gem(
        self,
        gem_instance,
        gem_class: typing.Type[CutterEngineBaseModel],
        field_name,
    ):
        for related_gem_side_effect in gem_class.CutterEngineMeta.related_gem_side_effects:
            query_field_name = f"{gem_settings.CUTTER_MODEL_RELATED_NAME}_id__in"
            query = {query_field_name: related_gem_side_effect.callback(gem_instance)}
            cut_related_gems = related_gem_side_effect.related_gem_model.objects.filter(**query)
            if field_name in related_gem_side_effect.self_affected_fields:
                for model_instance in cut_related_gems:
                    self.cut_model_fields(
                        model_instance.__getattribute__(gem_settings.CUTTER_MODEL_RELATED_NAME),
                        related_gem_side_effect.gem_fields,
                    )

    def cut_model(self, model_instance):
        autodiscover_modules("apps")
        cutter_model = self.cutter_registry.get_cutter_for_model(model_instance)
        if not cutter_model or not cutter_model.model:
            return None
        cut_fields = []
        for cutter in cutter_model.triggers:
            gem_class = cutter.gem_class
            if not gem_class:
                continue
            cut_fields.extend(gem_class.get_gem_fields())
        if not cut_fields:
            return None
        return self.cut_model_fields(model_instance, cut_fields)


forge = Forge(cutter_registry)



import typing

from django.db import transaction

from django_gem.constants import GemCuttingMode
from django_gem.entities.context import gem_cutting_context
from django_gem.entities.registry import CutterBatchItem
from django_gem.overrides.utils import add_cut_items

CUTTER_M2M_SIGNAL_ACTIONS = ["pre_add", "pre_remove", "pre_clear"]


def m2m_changed_hook(
    cut_batch_items: typing.List[CutterBatchItem],
):
    def m2m_changed(sender, instance, action, **kwargs):  # noqa
        from django_gem.toolkit import gem_settings, smith

        gem_cutting_enabled = gem_settings.GEM_CUTTING_ENABLED

        if action not in CUTTER_M2M_SIGNAL_ACTIONS:
            return

        with transaction.atomic():
            changed_cut_batch_items = []
            for cut_batch_item in cut_batch_items:
                cut_batch_item.load_object_ids(instance)
                changed_cut_batch_items.append(cut_batch_item)

            if not gem_cutting_enabled:
                return

            add_cut_items(smith, changed_cut_batch_items)

            if gem_cutting_context.cutting_mode == GemCuttingMode.TRANSACTION:
                transaction.on_commit(
                    lambda: smith.initiate_cutting(),
                )

    return m2m_changed



from celery import shared_task


@shared_task(name="cut_content_type_task")
def cut_content_type_task(natural_key: str, field_names: list):
    from django_gem.toolkit.saw import Saw

    Saw.cut_content_type(natural_key, field_names)



from celery import shared_task


@shared_task(name="cut_queryset_task")
def cut_queryset_task(natural_key: str, object_ids: list):
    from django_gem.toolkit.saw import Saw

    Saw.cut_queryset(natural_key, object_ids)



from django.test import override_settings

from django_gem.toolkit import gem_settings


class override_gem_settings(override_settings):  # noqa
    def enable(self):
        super().enable()
        gem_settings.load()

    def disable(self):
        super().disable()
        gem_settings.load()



import typing

from django.db import models
from django.db.models import Manager
from django.db.models.fields.related_descriptors import ManyToManyDescriptor

from django_gem.constants import CallbackType
from django_gem.models.base import CutterSideEffect
from django_gem.toolkit import gem_settings


def side_effects(
    *args: typing.Tuple[
        typing.Union[typing.Type[typing.Union[models.Model, ManyToManyDescriptor]], Manager],
        typing.List[str],
        CallbackType,
    ],
):
    def wrapped_function(function):
        if gem_settings.CUTTER_PROPERTY_PREFIX not in function.__name__:
            return function

        cutter_field_name = function.__name__.replace(gem_settings.CUTTER_PROPERTY_PREFIX, "")
        function.side_effects = [
            CutterSideEffect(
                model=model_class,
                callback=callback,
                affected_fields=affected_fields,
                gem_fields=[cutter_field_name],
            )
            for (model_class, affected_fields, callback) in args
        ]
        return function

    return wrapped_function



import typing
from collections import defaultdict

from django.db.models.fields.related_descriptors import ManyToManyDescriptor
from django.db.models.signals import m2m_changed
from django.utils.module_loading import autodiscover_modules

from django_gem.core import get_model_natural_key
from django_gem.entities.registry import (
    CutterBatchItem,
    CutterModel,
    CutterModelTrigger,
    EagerGemItem,
    ReverseCutterModel,
)
from django_gem.models.base import CutterEngineBaseModel, CutterSideEffect
from django_gem.overrides import (
    delete_cutting_hook,
    m2m_changed_hook,
    save_cutting_hook,
)
from django_gem.toolkit import gem_settings
from django_gem.validators import CutterEngineValidator


class CutterRegistry:
    """Collects models dependencies to build cutter classes dependencies map"""

    # Direct cutters contains the mapping of which model needs to be updated and how
    direct_cutters = defaultdict(CutterModel)
    # Reverse cutters contains the related models mapping that should trigger updates
    reverse_cutters = defaultdict(ReverseCutterModel)

    @classmethod
    def get_model_key(cls, model_class):
        if isinstance(model_class, ManyToManyDescriptor):
            return get_model_natural_key(model_class.through)
        return get_model_natural_key(model_class)  # noqa

    def get_cutter_for_model(self, model_class):
        autodiscover_modules("apps")
        return self.direct_cutters.get(self.get_model_key(model_class))

    def get_reverse_cutter_for_model(self, model_class):
        autodiscover_modules("apps")
        return self.reverse_cutters.get(self.get_model_key(model_class))

    def get_model_eager_gem_item(self, model_instance, field_name) -> typing.Optional[EagerGemItem]:
        cutter_model = self.get_cutter_for_model(model_instance)
        if not cutter_model or not cutter_model.model:
            return None
        for registered_cutter in cutter_model.triggers:
            gem_class = registered_cutter.gem_class
            if field_name not in [field.name for field in gem_class._meta.fields]:
                continue
            related_gem_field_name = (
                getattr(gem_class.CutterEngineMeta, "related_name", None)
                or gem_settings.DEFAULT_RELATED_GEM_FIELD_NAME
            )
            return EagerGemItem(
                related_gem_field_name=related_gem_field_name,
                cutter=registered_cutter.cutter,
            )
        return None

    def get_eager_field_value(self, model_instance, field_name, default_value=None):
        eager_gem_item = self.get_model_eager_gem_item(
            model_instance=model_instance,
            field_name=field_name,
        )
        if not eager_gem_item:
            return default_value
        cached_cutter_field_name = (
            f"__{eager_gem_item.related_gem_field_name}_{gem_settings.CACHED_GEM_CUTTER_FIELD_NAME}"
        )
        cutter = getattr(model_instance, cached_cutter_field_name, None)
        if not cutter:
            cutter = eager_gem_item.cutter(model_instance)
            setattr(model_instance, cached_cutter_field_name, cutter)
        if hasattr(cutter, f"{gem_settings.CUTTER_PROPERTY_PREFIX}{field_name}"):
            return getattr(cutter, f"{gem_settings.CUTTER_PROPERTY_PREFIX}{field_name}")
        return default_value

    def register_cutter(self, gem_class: typing.Type[CutterEngineBaseModel]):
        model_key = self.get_model_key(gem_class.CutterEngineMeta.model)
        gem_class.CutterEngineMeta.side_effects = [
            *gem_class.CutterEngineMeta.side_effects,
            *gem_class.CutterEngineMeta.cutter.side_effects,
        ]
        CutterEngineValidator(gem_class).validate()
        self.direct_cutters[model_key].model = gem_class.CutterEngineMeta.model
        self.direct_cutters[model_key].triggers.append(
            CutterModelTrigger(
                cutter=gem_class.CutterEngineMeta.cutter,
                gem_class=gem_class,
                propagate_triggers=gem_class.CutterEngineMeta.propagate_triggers,
            )
        )
        self.direct_cutters[model_key].select_related_fields = {
            *self.direct_cutters[model_key].select_related_fields,
            *gem_class.CutterEngineMeta.cut_select_related_fields,
        }
        self.direct_cutters[model_key].prefetch_related_fields = {
            *self.direct_cutters[model_key].prefetch_related_fields,
            *gem_class.CutterEngineMeta.cut_prefetch_related_fields,
        }
        for side_effect in gem_class.CutterEngineMeta.side_effects:
            self._add_side_effect_to_triggers(gem_class, side_effect)

    def _add_side_effect_to_triggers(
        self,
        gem_class: typing.Type[CutterEngineBaseModel],
        side_effect: CutterSideEffect,
    ):
        side_effect_model = side_effect.model
        side_effect_model_key = self.get_model_key(side_effect_model)
        gem_fields = getattr(side_effect, "gem_fields", None) or gem_class.get_gem_fields()
        side_effect_affected_fields = getattr(side_effect, "affected_fields", None)
        for trigger in self.reverse_cutters[side_effect_model_key].triggers:
            if (
                trigger.callback_group == side_effect.callback_group
                and (trigger.model == gem_class.CutterEngineMeta.model)
                and (trigger.affected_fields == side_effect_affected_fields)
            ):
                trigger.gem_fields = [
                    *trigger.gem_fields,
                    *gem_fields,
                ]
                return
        self.reverse_cutters[side_effect_model_key].model = side_effect_model
        self.reverse_cutters[side_effect_model_key].triggers.append(
            CutterSideEffect(
                model=gem_class.CutterEngineMeta.model,
                callback=side_effect.callback,
                gem_fields=gem_fields,
                affected_fields=side_effect_affected_fields,
            )
        )

    def load(self):
        autodiscover_modules("gems")
        for subclass in self.get_subclasses(CutterEngineBaseModel):
            self.register_cutter(subclass)

    def get_subclasses(
        self, parent_class: typing.Type[CutterEngineBaseModel]
    ) -> typing.List[typing.Type[CutterEngineBaseModel]]:
        subclasses: typing.List[typing.Type[CutterEngineBaseModel]] = []
        for subclass in parent_class.__subclasses__():
            if len(subclass.__subclasses__()) > 0:
                subclasses.extend(self.get_subclasses(subclass))
            else:
                subclasses.append(subclass)
        return subclasses

    def ready(self):
        self.load()
        for app, triggered_model in self.reverse_cutters.items():
            cut_batch_items = [
                CutterBatchItem(
                    model=item.model,
                    gem_fields=item.gem_fields,
                    affected_fields=item.affected_fields,
                    callback=item.callback,
                )
                for item in triggered_model.triggers
            ]
            if isinstance(triggered_model.model, ManyToManyDescriptor):
                m2m_changed.connect(
                    m2m_changed_hook(cut_batch_items),
                    sender=triggered_model.model.through,  # noqa
                )
            else:
                triggered_model.model.save = save_cutting_hook(
                    triggered_model.model.save,
                    cut_batch_items,
                )
                triggered_model.model.delete = delete_cutting_hook(
                    triggered_model.model.delete,
                    cut_batch_items,
                )


cutter_registry = CutterRegistry()



from django.conf import settings


class DjangoGemsSettings:
    # region Non-configurable fields
    DEFAULT_RELATED_GEM_FIELD_NAME = "gem"
    CACHED_GEM_CUTTER_FIELD_NAME = "cached_cutter"
    # endregion

    CUTTER_PROPERTY_PREFIX = "cut_"
    CUTTER_MODEL_RELATED_NAME = "cutter_target"
    CUTTER_PROPAGATED_TRIGGERS_MAX_DEPTH = 10
    GEM_IGNORED_FIELDS = []
    GEM_CUTTING_ENABLED = True
    GEM_ANVILS = []

    def __init__(self):
        self.load()

    def load(self):
        if hasattr(settings, "CUTTER_PROPERTY_PREFIX"):
            self.CUTTER_PROPERTY_PREFIX = settings.CUTTER_PROPERTY_PREFIX
        if hasattr(settings, "CUTTER_PROPAGATED_TRIGGERS_MAX_DEPTH"):
            self.CUTTER_PROPAGATED_TRIGGERS_MAX_DEPTH = (
                settings.CUTTER_PROPAGATED_TRIGGERS_MAX_DEPTH
            )
        if hasattr(settings, "CUTTER_MODEL_RELATED_NAME"):
            self.CUTTER_MODEL_RELATED_NAME = settings.CUTTER_MODEL_RELATED_NAME
        if hasattr(settings, "GEM_IGNORED_FIELDS"):
            self.GEM_IGNORED_FIELDS = settings.GEM_IGNORED_FIELDS
        if hasattr(settings, "GEM_CUTTING_ENABLED"):
            self.GEM_CUTTING_ENABLED = settings.GEM_CUTTING_ENABLED
        if hasattr(settings, "GEM_ANVILS"):
            self.GEM_ANVILS = settings.GEM_ANVILS


gem_settings = DjangoGemsSettings()



from django.apps import AppConfig


class DjangoGemsConfig(AppConfig):
    name = "django_gem"

    def ready(self):
        from django_gem.toolkit import cutter_registry, smith

        cutter_registry.ready()
        smith.load()
        super().ready()



import json

from django_gem.toolkit.anvils.base import BaseAnvil


class EagerAnvil(BaseAnvil):
    @classmethod
    def cut(cls, sealed_chest, sealed_sources):
        from django_gem.toolkit.saw import Saw

        Saw.cut_models(
            sealed_chest=json.dumps(sealed_chest), sealed_sources=json.dumps(sealed_sources)
        )



from django_gem.context import CuttingAlwaysEager


def cutting_always_eager(func):
    def inner_func(*args, **kwargs):
        with CuttingAlwaysEager():
            return func(*args, **kwargs)

    return inner_func



import typing

from django.db import models


def get_model_natural_key(model: typing.Union[models.Model, typing.Type[models.Model]]):
    model_meta = model._meta
    return f"{model_meta.app_label}.{model_meta.model_name}"



import typing

from django_gem.core import get_model_natural_key
from django_gem.entities.registry import CutterBatchItem
from django_gem.models.mixins import GemTriggerMixin


def is_cutting_needed(model_instance, affected_fields):
    return (
        model_instance._state.adding  # noqa
        or not affected_fields
        or not isinstance(model_instance, GemTriggerMixin)
        or model_instance.is_any_field_changed(affected_fields)
    )


def add_cut_items(
    smith,
    cut_batch_items: typing.List[CutterBatchItem],
):
    # Marking cut as started outside the transaction to avoid locks
    for cut_batch_item in cut_batch_items:
        if not cut_batch_item.object_ids:
            continue

        smith.add_item(
            natural_key=get_model_natural_key(cut_batch_item.model),
            object_ids=list(set(cut_batch_item.object_ids)),
            gem_fields=list(set(cut_batch_item.gem_fields)),
        )



import typing
from dataclasses import dataclass, field

from django.db import models

from django_gem.constants import CallbackType
from django_gem.cutters.base import BaseCutter
from django_gem.models.base import CutterEngineBaseModel, CutterSideEffect

# region Configuration

# region Regular cutter entities


@dataclass
class CutterModelTrigger:
    cutter: typing.Type[BaseCutter]
    gem_class: typing.Type[CutterEngineBaseModel]
    propagate_triggers: typing.Dict[str, typing.List] = field(default_factory=lambda: {})


@dataclass
class CutterModel:
    model: typing.Type[models.Model] = None
    triggers: typing.List[CutterModelTrigger] = field(default_factory=lambda: [])
    select_related_fields: typing.List[str] = field(default_factory=lambda: [])
    prefetch_related_fields: typing.List[str] = field(default_factory=lambda: [])


# endregion

# region Reverse cutter entities


@dataclass
class ReverseCutterModel:
    model: typing.Type[models.Model] = None
    triggers: typing.List[CutterSideEffect] = field(default_factory=lambda: [])


# endregion

# endregion


# region Batches
@dataclass
class CutterBatchItem:
    model: typing.Type[models.Model]
    callback: CallbackType
    gem_fields: typing.List[str]
    affected_fields: typing.List[str] = field(default_factory=lambda: [])
    object_ids: typing.List[str] = field(default_factory=lambda: [])

    def load_object_ids(self, instance):
        try:
            self.object_ids = list(self.callback(instance))
        except Exception:  # noqa
            ...


# endregion


# region Eager Gem Recalculations
@dataclass
class EagerGemItem:
    cutter: typing.Type[BaseCutter]
    related_gem_field_name: str


# endregion



import typing

from django.db import models

from django_gem.entities.context import gem_cutting_context
from django_gem.models.mixins import GemModelMixin
from django_gem.toolkit import cutter_registry


class GemProperty:  # noqa: N801
    def __init__(self, func=None, default_value=None, override_gem_field_name=None):
        self.func = func
        if self.func:
            self.name = func.__name__
        if override_gem_field_name:
            self.name = override_gem_field_name
        self.default_value = default_value

    def __call__(self, *args, **kwargs):
        if len(args) != 1:
            raise ValueError("`gem_property` is misconfigured", args, kwargs)
        self.func = args[0]
        self.name = self.func.__name__
        return self

    def __get__(
        self, instance: typing.Union[typing.Optional[GemModelMixin], models.Model], cls=None
    ):
        if not instance:
            return self.func

        should_return_eager_value = gem_cutting_context.return_eager_value
        if isinstance(instance, GemModelMixin):
            should_return_eager_value = (
                should_return_eager_value and not instance.is_cutting_in_progress
            )
        try:
            if not should_return_eager_value:
                return self.func(instance)
        except AttributeError:
            ...
        return cutter_registry.get_eager_field_value(
            instance,
            self.name,
            default_value=self.default_value,
        )


gem_property = GemProperty



from django.contrib.contenttypes.models import ContentType
from django.db import models

from django_gem.constants import GemLogEntryAction


class GemLogEntryManager(models.Manager):
    def log_create(self, instance, **kwargs):
        changes = kwargs.get("changes", None)
        pk = self._get_pk_value(instance)

        content_type = ContentType.objects.get_for_model(instance)

        if changes is not None:
            kwargs.setdefault("content_type", content_type)
            kwargs.setdefault("object_id", pk)

            if kwargs.get("action", None) is GemLogEntryAction.CREATE:
                self.filter(content_type=content_type, object_id=pk).delete()
            return self.create(**kwargs)
        return None

    def _get_pk_value(self, instance):
        """
        Get the primary key field value for a model instance.

        :param instance: The model instance to get the primary key for.
        :type instance: Model
        :return: The primary key value of the given model instance.
        """
        pk_field = instance._meta.pk.name  # noqa
        pk = getattr(instance, pk_field, None)

        # Check to make sure that we got a pk not a model object.
        if isinstance(pk, models.Model):
            pk = self._get_pk_value(pk)
        return pk



import typing

from django_gem.exceptions import (
    CutterClassFieldMissingException,
    CutterFieldMissingException,
    UnusedCutterMethodException,
)
from django_gem.logger import logger
from django_gem.models.base import CutterEngineBaseModel, CutterSideEffect
from django_gem.toolkit import gem_settings


class CutterEngineValidator:
    def __init__(self, gem_class: typing.Type[CutterEngineBaseModel]):
        self.gem_class = gem_class
        self.model = gem_class.CutterEngineMeta.model
        self.self_affected_fields = gem_class.CutterEngineMeta.self_affected_fields
        self.propagate_triggers = gem_class.CutterEngineMeta.propagate_triggers
        self.cutter = gem_class.CutterEngineMeta.cutter
        self.side_effects = gem_class.CutterEngineMeta.side_effects

    def validate_cutter_methods(self):
        for cutter_method_name in self.cutter.__dict__.keys():
            if not cutter_method_name.startswith(gem_settings.CUTTER_PROPERTY_PREFIX):
                continue
            field_name = cutter_method_name[len(gem_settings.CUTTER_PROPERTY_PREFIX) :]  # noqa
            if not hasattr(self.gem_class, field_name):
                exception = UnusedCutterMethodException(
                    cutter_class=self.cutter,
                    field_name=field_name,
                )
                logger.warning(exception)

    def validate_self_fields(self):
        for self_affected_field in self.self_affected_fields:
            if not hasattr(self.model, self_affected_field):
                raise CutterFieldMissingException(
                    model_class=self.model,
                    field_name=self_affected_field,
                )
        for propagate_trigger_key, propagate_trigger_values in self.propagate_triggers.items():
            if not hasattr(self.gem_class, propagate_trigger_key):
                raise CutterFieldMissingException(
                    model_class=self.gem_class,
                    field_name=propagate_trigger_key,
                )
            if not hasattr(
                self.cutter, f"{gem_settings.CUTTER_PROPERTY_PREFIX}{propagate_trigger_key}"
            ):
                raise CutterClassFieldMissingException(
                    cutter_class=self.cutter,
                    field_name=propagate_trigger_key,
                )
            for propagate_trigger_value in propagate_trigger_values:
                if not hasattr(self.gem_class, propagate_trigger_value):
                    raise CutterFieldMissingException(
                        model_class=self.gem_class,
                        field_name=propagate_trigger_value,
                    )
                if not hasattr(
                    self.cutter, f"{gem_settings.CUTTER_PROPERTY_PREFIX}{propagate_trigger_value}"
                ):
                    raise CutterClassFieldMissingException(
                        cutter_class=self.cutter,
                        field_name=propagate_trigger_value,
                    )

    def validate_side_effect_fields(self, side_effect: CutterSideEffect):
        side_effect_model = side_effect.model
        for affected_field in side_effect.affected_fields:
            if not hasattr(side_effect_model, affected_field):
                raise CutterFieldMissingException(
                    model_class=side_effect_model,
                    field_name=affected_field,
                )
        if not side_effect.gem_fields:
            return
        for gem_field in side_effect.gem_fields:
            if not hasattr(self.gem_class, gem_field):
                raise CutterFieldMissingException(
                    model_class=self.gem_class,
                    field_name=gem_field,
                )
            if not hasattr(self.cutter, f"{gem_settings.CUTTER_PROPERTY_PREFIX}{gem_field}"):
                raise CutterClassFieldMissingException(
                    cutter_class=self.cutter,
                    field_name=gem_field,
                )

    def validate(self):
        self.validate_self_fields()
        self.validate_cutter_methods()
        for side_effect in self.side_effects:
            self.validate_side_effect_fields(side_effect)



import logging

from django.contrib.auth import get_user_model
from django.core.management import BaseCommand

from django_gem.toolkit.saw import Saw

User = get_user_model()
logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "app_label",
            help="Specify the app label to cut gems for.",
        )
        parser.add_argument(
            "model_name",
            help="Specify the mode name to cut gems for.",
        )
        parser.add_argument(
            "--field-names",
            nargs="*",
            help="Specify the field names to cut gems for.",
        )

    def handle(self, *args, **options):
        app_label = options["app_label"]
        model_name = options["model_name"]
        field_names = options.get("field_names") or []
        Saw.cut_content_type(natural_key=f"{app_label}.{model_name}", field_names=field_names)



from django_gem.constants import GemCuttingMode


class GemCuttingContext:
    return_eager_value: bool = False
    cutting_mode: str = GemCuttingMode.TRANSACTION


gem_cutting_context = GemCuttingContext()



import typing

from django.db import models

from django_gem.cutters.base import BaseCutter


class CutterFieldMissingException(Exception):
    def __init__(self, model_class: typing.Type[models.Model], field_name: str):
        self.model_class_name = model_class.__name__
        self.field_name = field_name

    def __str__(self):
        return f'Can not find field "{self.field_name}" for model {self.model_class_name}'


class UnusedCutterMethodException(Exception):
    def __init__(self, cutter_class: typing.Type[BaseCutter], field_name: str):
        self.cutter_class_name = cutter_class.__name__
        self.field_name = field_name

    def __str__(self):
        return f'Unused cutter method "{self.field_name}" for cutter {self.cutter_class_name}'


class CutterClassFieldMissingException(Exception):
    def __init__(self, cutter_class: typing.Type[BaseCutter], field_name: str):
        self.cutter_class_name = cutter_class.__name__
        self.field_name = field_name

    def __str__(self):
        return f'Can not find method "{self.field_name}" for cutter {self.cutter_class_name}'



# Generated by Django 4.2.1 on 2023-05-07 17:15

import uuid

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="GemLogEntry",
            fields=[
                (
                    "id",
                    models.UUIDField(
                        default=uuid.uuid4,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                (
                    "object_id",
                    models.CharField(db_index=True, max_length=255, verbose_name="object id"),
                ),
                (
                    "action",
                    models.PositiveSmallIntegerField(
                        choices=[(0, "create"), (1, "update"), (2, "delete")],
                        db_index=True,
                        verbose_name="action",
                    ),
                ),
                ("changes", models.JSONField(verbose_name="Model changed fields")),
                (
                    "timestamp",
                    models.DateTimeField(
                        auto_now_add=True, db_index=True, verbose_name="timestamp"
                    ),
                ),
                (
                    "actor",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="actor",
                    ),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="contenttypes.contenttype",
                        verbose_name="content type",
                    ),
                ),
            ],
            options={
                "verbose_name": "gem log entry",
                "verbose_name_plural": "gem log entries",
                "ordering": ["-timestamp"],
                "get_latest_by": "timestamp",
            },
        ),
    ]



import logging

from django.contrib.auth import get_user_model
from django.core.management import BaseCommand

from django_gem.toolkit.saw import Saw

User = get_user_model()
logger = logging.getLogger(__name__)


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "app_label",
            help="Specify the app label to cut gems for.",
        )
        parser.add_argument(
            "model_name",
            help="Specify the mode name to cut gems for.",
        )
        parser.add_argument(
            "object_id",
            help="Specify the object id to cut gems for.",
        )

    def handle(self, *args, **options):
        app_label = options["app_label"]
        model_name = options["model_name"]
        object_id = options["object_id"]
        Saw.cut_queryset(f"{app_label}.{model_name}", object_ids=[object_id])



import typing

from django.db import models
from django.db.models.base import ModelBase

from django_gem.entities.models import CutterRelatedGemSideEffect, CutterSideEffect
from django_gem.toolkit.settings import gem_settings


class CutterEngineMetaBase:
    model = None
    related_name = None
    cutter = None
    self_affected_fields: typing.List[str] = []
    propagate_triggers: typing.Dict[str, typing.List] = {}
    side_effects: typing.List[CutterSideEffect] = []
    related_gem_side_effects: typing.List[CutterRelatedGemSideEffect] = []
    cut_select_related_fields: typing.List[str] = []
    cut_prefetch_related_fields: typing.List[str] = []


class CutterEngineBase(ModelBase):
    def __new__(cls, name, bases, attrs):
        if cutter_engine_meta := attrs.get("CutterEngineMeta"):
            if cutter_engine_meta and cutter_engine_meta.model:
                related_name = (
                    getattr(cutter_engine_meta, "related_name", None)
                    or gem_settings.DEFAULT_RELATED_GEM_FIELD_NAME
                )
                attrs[gem_settings.CUTTER_MODEL_RELATED_NAME] = models.OneToOneField(
                    cutter_engine_meta.model,
                    on_delete=models.CASCADE,
                    null=True,
                    related_name=related_name,
                )
        klass = super().__new__(cls, name, bases, attrs)
        return klass


class CutterEngineBaseModel(models.Model, metaclass=CutterEngineBase):
    BASE_FIELDS = ["id", gem_settings.CUTTER_MODEL_RELATED_NAME]

    def __str__(self):
        return f"Gem of {self.__getattribute__(gem_settings.CUTTER_MODEL_RELATED_NAME)}"

    class Meta:
        abstract = True

    @classmethod
    def get_gem_fields(cls):
        ignored_fields = cls.BASE_FIELDS
        if gem_ignored_fields := gem_settings.GEM_IGNORED_FIELDS:
            ignored_fields = [*cls.BASE_FIELDS, *gem_ignored_fields]
        return [field.name for field in cls._meta.fields if field.name not in ignored_fields]

    class CutterEngineMeta(CutterEngineMetaBase):
        ...
